import os
from string import Template

import pytest

import tests.t_utils.common as t_common


@pytest.mark.parametrize(
    ("query", "expect"),
    [
        ("select ?name where {ex:Picasso foaf:firstName ?name}", [{"name": "Pablo"}]),
        (
            "select ?city {ex:Picasso ex:homeAddress ?address . ?address ex:city ?city}",
            [{"city": "Madrid"}],
        ),
    ],
)
def test_rdf4j_picasso_data(query: str, expect: list[dict[str, str]]):
    try:
        client = t_common.initialized_rdf4j_repo()
    except Exception as exc:
        if os.getenv("CI"):
            pytest.fail(f"{exc}")
        else:
            pytest.skip(f"{exc}")

    prefixes = Template("PREFIX ex:<${ex}>\nPREFIX foaf:<${foaf}>").substitute(client.prefixes)
    result = client.exec_query(f"{prefixes}\n{query}").results.values_as_dict()
    assert result == expect



from __future__ import annotations

import datetime
from decimal import Decimal
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd
import pandera.pandas as pa
import pytest
from pandas.testing import assert_frame_equal
from pandera.typing import DataFrame

from cimsparql import type_mapper
from cimsparql.data_models import CoercingSchema
from cimsparql.graphdb import RestApi
from cimsparql.model import ServiceConfig

if TYPE_CHECKING:
    from typing import Any

    from pytest_httpserver import HTTPServer


def init_triple_store_server(httpserver: HTTPServer, sparql_result: dict[str, Any] | None = None) -> ServiceConfig:
    """Create a triple store server that returns sparql_result_data when a call is made."""
    sparql_result = sparql_result or empty_sparql_result()
    httpserver.expect_request("/sparql").respond_with_json(sparql_result)
    return ServiceConfig(server=httpserver.url_for("/sparql"), rest_api=RestApi.DIRECT_SPARQL_ENDPOINT)


def empty_sparql_result() -> dict[str, Any]:
    return {"head": {"vars": []}, "results": {"bindings": []}}


def test_python_type_map_bool():
    type_converter = type_mapper.XSD_TYPE_MAP["boolean"]
    assert callable(type_converter)
    assert type_converter("TRUE")
    assert not type_converter("FALSE")


def test_get_map_empty_pandas(httpserver: HTTPServer):
    service_cfg = init_triple_store_server(httpserver)
    mapper = type_mapper.TypeMapper(service_cfg)
    assert mapper.get_map() == {}


def test_map_data_types(httpserver: HTTPServer):
    results = {
        "sparql_type": ["http://c#Degrees", "http://c#Status", "http://c#Amount"],
        "range": ["http://x#Float", "http://x#Bool", "http://x#Integer"],
    }

    # Represent result as a sparql_result
    sparql_result = {
        "head": {"vars": list(results.keys())},
        "results": {
            "bindings": [{k: {"type": "literal", "value": v[i]} for k, v in results.items()} for i in range(3)]
        },
    }
    service_cfg = init_triple_store_server(httpserver, sparql_result)

    types = {"http://c#Degrees": float, "http://c#Status": bool, "http://c#Amount": int}
    tm = type_mapper.TypeMapper(service_cfg, custom_additions=types)

    angles = pd.DataFrame(
        {
            "angle": ["1.0", "2.0", "3.0"],
            "active": ["true", "true", "false"],
            "number": ["1", "2", "3"],
        }
    )

    expect_angles = angles.astype({"angle": float, "active": bool, "number": int})

    col_map = {
        "angle": "http://c#Degrees",
        "active": "http://c#Status",
        "number": "http://c#Amount",
    }
    result = tm.map_data_types(angles, col_map)

    assert_frame_equal(result, expect_angles)


@pytest.mark.parametrize(
    ("dtype", "test"),
    [
        ("boolean", ("true", True)),
        ("boolean", ("false", False)),
        ("boolean", ("1", True)),
        ("boolean", ("0", False)),
        ("date", ("2021-12-14", pd.Timestamp("2021-12-14 00:00:00"))),
        (
            "dateTime",
            ("2002-10-14T12:00:00", datetime.datetime(2002, 10, 14, 12, 0, 0)),  # noqa: DTZ001
        ),
        ("dateTime", ("2002-10-14T12:00:00Z", pd.Timestamp("2002-10-14 12:00:00+0000", tz="UTC"))),
        ("duration", ("P365D", datetime.timedelta(365))),
        ("duration", ("-P365D", datetime.timedelta(-365))),
        (
            "duration",
            ("P3DT5H20M30.123S", datetime.timedelta(days=3, seconds=5 * 3600 + 20 * 60 + 30.123)),
        ),
        ("decimal", ("2.3", Decimal("2.3"))),
        ("time", ("21:32:52", pd.Timestamp("21:32:52"))),
        ("time", ("21:32:52Z", pd.Timestamp("21:32:52", tz="UTC"))),
    ],
)
def test_xsd_types(dtype: str, test: tuple[str, Any]):
    value, expect = test
    type_converter = type_mapper.XSD_TYPE_MAP[dtype]
    assert callable(type_converter)
    assert type_converter(value) == expect


@pytest.mark.parametrize(
    ("cim_version", "have_cim_version"),
    [
        ("10", False),
        # Because of the default prefixes, 16 is default. Thus, it should be present
        ("16", True),
    ],
)
def test_have_cim_version(httpserver: HTTPServer, cim_version: int, have_cim_version: bool):
    service_cfg = init_triple_store_server(httpserver)
    tm = type_mapper.TypeMapper(service_cfg)
    assert tm.have_cim_version(str(cim_version)) == have_cim_version


class FloatSchema(CoercingSchema):
    float_col: float = pa.Field(nullable=True)


FloadDataFrame = DataFrame[FloatSchema]


@pytest.mark.parametrize(
    ("in_data", "out_data"),
    [
        ({"float_col": ["1.0", "2.0"]}, {"float_col": [1.0, 2.0]}),
        ({"float_col": ["1.0", None]}, {"float_col": [1.0, np.nan]}),
    ],
)
def test_coerce_missing_type(httpserver: HTTPServer, in_data: dict[str, Any], out_data: dict[str, Any]):
    client = init_triple_store_server(httpserver)
    in_df, out_df = pd.DataFrame(in_data), pd.DataFrame(out_data)
    mapper = type_mapper.TypeMapper(client)

    col_map = {"float_col": "http://non-existent#type"}
    df = FloadDataFrame(mapper.map_data_types(in_df, col_map))
    pd.testing.assert_frame_equal(df, out_df)


def test_coerce_float_from_literal(httpserver: HTTPServer):
    client = init_triple_store_server(httpserver)
    mapper = type_mapper.TypeMapper(client)

    # Set the type to literal which could happen if the triple store contains non-typed entries.
    # The panders schemas should still be able to cast into the correct type
    col_map = {"float_col": "literal"}
    df = pd.DataFrame({"float_col": ["1.0", None]})

    # None should now be preserved by the mapper
    df = mapper.map_data_types(df, col_map)
    assert df.loc[1, "float_col"] is None

    df = FloadDataFrame(df)

    # Now the None value should be casted into np.nan
    missing = df.loc[1, "float_col"]
    assert missing is not None
    assert pd.isna(missing)



import asyncio
import logging
import os
import re
from base64 import b64encode
from collections.abc import Callable
from concurrent.futures import ThreadPoolExecutor
from dataclasses import asdict, dataclass
from functools import lru_cache
from http import HTTPStatus
from typing import Any, ParamSpec, TypeVar

import httpx
import pandas as pd
import pytest
from pytest_httpserver import HeaderValueMatcher, HTTPServer
from SPARQLWrapper import SPARQLWrapper

import cimsparql.data_models as cim_dm
import tests.t_utils.common as t_common
import tests.t_utils.custom_models as t_custom
from cimsparql.graphdb import (
    GraphDBClient,
    RepoInfo,
    RestApi,
    ServiceConfig,
    config_bytes_from_template,
    confpath,
    data_row,
    new_repo,
    repos,
)
from cimsparql.model import Model, SingleClientModel
from cimsparql.sparql_result_json import SparqlResultJsonFactory, SparqlResultValue
from cimsparql.type_mapper import TypeMapper

logger = logging.getLogger()

T = TypeVar("T")
P = ParamSpec("P")
R = TypeVar("R")


def exception_logging(func: Callable[P, R], *args: P.args, **kwargs: P.kwargs) -> R:
    logger.debug("Starting %s", func.__name__)
    try:
        return func(*args, **kwargs)
    except Exception:
        logger.exception("Exception for %s", func.__name__)
        return R() if callable(R) else R


@dataclass
class DataResult:
    ac_lines: cim_dm.AcLinesDataFrame
    associated_switches: cim_dm.AssociatedSwitchesDataFrame
    base_voltage: cim_dm.BaseVoltageDataFrame
    borders: cim_dm.BordersDataFrame
    bus_data: cim_dm.BusDataFrame
    connections: cim_dm.ConnectionsDataFrame
    connectivity_nodes: cim_dm.ConnectivityNodeDataFrame
    converters: cim_dm.ConvertersDataFrame
    coordinates: cim_dm.CoordinatesDataFrame
    dc_active_flow: cim_dm.DcActiveFlowDataFrame
    disconnected: cim_dm.DisconnectedDataFrame
    exchange: cim_dm.ExchangeDataFrame
    full_model: cim_dm.FullModelDataFrame
    gen_unit_and_sync_machine_mrid: cim_dm.GenUnitAndSyncMachineMridDataFrame
    hvdc_converter_bidzones: cim_dm.HVDCBidzonesDataFrame
    loads: cim_dm.LoadsDataFrame
    market_dates: cim_dm.MarketDatesDataFrame
    phase_tap_changer: cim_dm.PhaseTapChangerDataFrame
    powerflow: cim_dm.PowerFlowDataFrame
    regions: cim_dm.RegionsDataFrame
    series_compensators: cim_dm.BranchComponentDataFrame
    station_group_codes_and_names: cim_dm.StationGroupCodeNameDataFrame
    substation_voltage_level: cim_dm.SubstationVoltageDataFrame
    station_group_for_power_unit: cim_dm.StationGroupForPowerUnitDataFrame
    sv_power_deviation: cim_dm.SvPowerDeviationDataFrame
    switches: cim_dm.SwitchesDataFrame
    synchronous_machines: cim_dm.SynchronousMachinesDataFrame
    transformer_branches: cim_dm.TransformerWindingDataFrame
    transformer_windings: cim_dm.TransformerWindingsDataFrame
    transformers: cim_dm.TransformersDataFrame
    transformers_connected_to_converter: cim_dm.TransfConToConverterDataFrame
    transformer_center_nodes: cim_dm.BusDataFrame
    wind_generating_units: cim_dm.WindGeneratingUnitsDataFrame


@lru_cache
def all_data(model: Model) -> DataResult:
    return asyncio.run(collect_data(model))


async def collect_data(model: Model) -> DataResult:
    loop = asyncio.get_event_loop()

    with ThreadPoolExecutor() as executor:
        ac_line_future = loop.run_in_executor(executor, model.ac_lines)
        associated_switches_future = loop.run_in_executor(executor, model.associated_switches)
        base_voltage_future = loop.run_in_executor(executor, model.base_voltage)
        borders_future = loop.run_in_executor(executor, model.borders)
        bus_data_future = loop.run_in_executor(executor, model.bus_data)
        connections_future = loop.run_in_executor(executor, model.connections)
        connectivity_nodes_future = loop.run_in_executor(executor, model.connectivity_nodes)
        converters_future = loop.run_in_executor(executor, model.converters)
        coordinates_future = loop.run_in_executor(executor, model.coordinates)
        dc_active_flow_future = loop.run_in_executor(executor, model.dc_active_flow)
        disconnected_future = loop.run_in_executor(executor, model.disconnected)
        exchange_future = loop.run_in_executor(executor, model.exchange)
        full_model_future = loop.run_in_executor(executor, model.full_model)
        gen_unit_and_sync_machine_mrid_future = loop.run_in_executor(executor, model.gen_unit_and_sync_machine_mrid)
        hvdc_converter_bidzones_future = loop.run_in_executor(executor, model.hvdc_converter_bidzones)
        loads_future = loop.run_in_executor(executor, model.loads)
        market_dates_future = loop.run_in_executor(executor, model.market_dates)
        phase_tap_changer_future = loop.run_in_executor(executor, model.phase_tap_changer)
        powerflow_future = loop.run_in_executor(executor, model.powerflow)
        regions_future = loop.run_in_executor(executor, model.regions)
        series_compensators_future = loop.run_in_executor(executor, model.series_compensators)
        station_group_codes_and_names_future = loop.run_in_executor(executor, model.station_group_codes_and_names)
        substation_voltage_level_future = loop.run_in_executor(executor, model.substation_voltage_level)
        station_group_for_power_unit_future = loop.run_in_executor(executor, model.station_group_for_power_unit)
        sv_power_deviation_future = loop.run_in_executor(executor, model.sv_power_deviation)
        switches_future = loop.run_in_executor(executor, model.switches)
        synchronous_machines_future = loop.run_in_executor(executor, model.synchronous_machines)
        transformer_branches_future = loop.run_in_executor(executor, model.transformer_branches)
        transformer_windings_future = loop.run_in_executor(executor, model.transformer_windings)
        transformers_future = loop.run_in_executor(executor, model.transformers)
        transformers_connected_to_converter_future = loop.run_in_executor(
            executor, model.transformers_connected_to_converter
        )
        transformer_center_nodes_future = loop.run_in_executor(executor, model.transformer_center_nodes)
        wind_generating_units_future = loop.run_in_executor(executor, model.wind_generating_units)

    await asyncio.wait(
        [
            ac_line_future,
            associated_switches_future,
            base_voltage_future,
            borders_future,
            bus_data_future,
            connections_future,
            connectivity_nodes_future,
            converters_future,
            coordinates_future,
            dc_active_flow_future,
            disconnected_future,
            exchange_future,
            full_model_future,
            gen_unit_and_sync_machine_mrid_future,
            hvdc_converter_bidzones_future,
            loads_future,
            market_dates_future,
            phase_tap_changer_future,
            powerflow_future,
            regions_future,
            series_compensators_future,
            station_group_codes_and_names_future,
            substation_voltage_level_future,
            station_group_for_power_unit_future,
            sv_power_deviation_future,
            switches_future,
            synchronous_machines_future,
            transformer_branches_future,
            transformer_windings_future,
            transformers_future,
            transformers_connected_to_converter_future,
            transformer_center_nodes_future,
            wind_generating_units_future,
        ]
    )
    return DataResult(
        ac_lines=ac_line_future.result(),
        associated_switches=associated_switches_future.result(),
        base_voltage=base_voltage_future.result(),
        borders=borders_future.result(),
        bus_data=bus_data_future.result(),
        connections=connections_future.result(),
        connectivity_nodes=connectivity_nodes_future.result(),
        converters=converters_future.result(),
        coordinates=coordinates_future.result(),
        dc_active_flow=dc_active_flow_future.result(),
        disconnected=disconnected_future.result(),
        exchange=exchange_future.result(),
        full_model=full_model_future.result(),
        gen_unit_and_sync_machine_mrid=gen_unit_and_sync_machine_mrid_future.result(),
        hvdc_converter_bidzones=hvdc_converter_bidzones_future.result(),
        loads=loads_future.result(),
        market_dates=market_dates_future.result(),
        phase_tap_changer=phase_tap_changer_future.result(),
        powerflow=powerflow_future.result(),
        regions=regions_future.result(),
        series_compensators=series_compensators_future.result(),
        station_group_codes_and_names=station_group_codes_and_names_future.result(),
        substation_voltage_level=substation_voltage_level_future.result(),
        station_group_for_power_unit=station_group_for_power_unit_future.result(),
        sv_power_deviation=sv_power_deviation_future.result(),
        switches=switches_future.result(),
        synchronous_machines=synchronous_machines_future.result(),
        transformer_branches=transformer_branches_future.result(),
        transformer_windings=transformer_windings_future.result(),
        transformers=transformers_future.result(),
        transformers_connected_to_converter=transformers_connected_to_converter_future.result(),
        transformer_center_nodes=transformer_center_nodes_future.result(),
        wind_generating_units=wind_generating_units_future.result(),
    )


@pytest.mark.parametrize("test_model", t_custom.all_custom_models())
def test_not_empty(test_model: t_common.ModelTest):
    model = test_model.model
    if not model:
        pytest.skip("Require access to GraphDB")
    dfs = all_data(model)

    for name, df in asdict(dfs).items():
        assert not df.empty, f"Failed for dataframe {name}"


@pytest.fixture
def model() -> Model:
    test_model = t_custom.federated_model()
    if not test_model.model:
        pytest.skip("Require access to GraphDB")
    return test_model.model


def test_cimversion(model: SingleClientModel):
    assert model.cim_version == 16


def test_ltc_fixed_angle_equals_zero(model: SingleClientModel):
    branches = model.get_table_and_convert(model.transformer_branches_query(), index="mrid")
    angle = model.winding_angle()
    pytest.approx(branches.loc[angle.index, "angle"] == 0.0)


def test_regions(model: SingleClientModel):
    regions = model.regions()
    assert pd.to_numeric(regions.groupby("region").count().loc["NO", "name"]) > 16


def test_hvdc_converters_bidzones(model: SingleClientModel):
    hvdc_converter_bidzone = model.hvdc_converter_bidzones()

    corridors = set[tuple[str, str]](
        zip(hvdc_converter_bidzone["bidzone_1"], hvdc_converter_bidzone["bidzone_2"], strict=True)
    )

    # Check data quality in the models
    expect_corridors = {("SE4", "SE3"), ("NO2", "DE"), ("NO2", "DK1"), ("NO2", "GB"), ("NO2", "NL")}
    assert expect_corridors.issubset(corridors)


def test_windings(model: SingleClientModel):
    windings = model.transformers(region="NO")
    assert windings.shape[1] == 9


@pytest.mark.skipif(os.getenv("GRAPHDB_SERVER") is None, reason="Need graphdb server to run")
def test_borders_no(model: SingleClientModel):
    borders = model.borders(region="NO")
    assert (borders[["area_1", "area_2"]] == "NO").any(axis=1).all()
    assert (borders["area_1"] != borders["area_2"]).all()


def sparql_result_int(num: int) -> SparqlResultValue:
    return SparqlResultValue(
        type="literal",
        datatype="http://www.w3.org/2001/XMLSchema#integer",
        value=str(num),
    )


def test_data_row():
    cols = ["a", "b", "c", "d", "e"]
    rows = [
        {"a": sparql_result_int(1), "b": sparql_result_int(2)},
        {"c": sparql_result_int(3), "d": sparql_result_int(4)},
        {"a": sparql_result_int(5), "b": sparql_result_int(6)},
        {"e": sparql_result_int(7)},
    ]
    assert not set(data_row(cols, rows)).symmetric_difference(cols)


def test_data_row_missing_column():
    cols = ["a", "b", "c", "d", "e"]
    rows = [
        {"a": sparql_result_int(1), "b": sparql_result_int(2)},
        {"c": sparql_result_int(3)},
        {"a": sparql_result_int(5), "b": sparql_result_int(6)},
        {"e": sparql_result_int(7)},
    ]
    assert set(data_row(cols, rows).keys()).symmetric_difference(cols) == {"d"}


def test_dtypes(model: SingleClientModel):
    mapper = TypeMapper(model.client.service_cfg)
    data = model.client.get_table(mapper.query)[0]
    assert data["sparql_type"].isna().sum() == 0


def test_prefix_resp_not_ok():
    resp = httpx.Response(status_code=HTTPStatus.UNAUTHORIZED, text="Something went wrong")
    with pytest.raises(RuntimeError) as exc:
        GraphDBClient(ServiceConfig(server="some-serever")).get_prefixes(
            http_transport=httpx.MockTransport(lambda _: resp)
        )
    assert resp.reason_phrase in str(exc)
    assert f"{resp.status_code}" in str(exc)


def test_conf_bytes_from_template():
    template = confpath() / "native_store_config_template.ttl"

    conf_bytes = config_bytes_from_template(template, {"repo": "test_repo"})
    conf_str = conf_bytes.decode()
    assert 'rep:repositoryID "test_repo"' in conf_str


def test_create_delete_repo():
    url = t_common.rdf4j_url()

    # Check if it is possible to make contact
    try:
        resp = httpx.get("http://" + url, timeout=5.0)
        if resp.status_code not in [HTTPStatus.OK, HTTPStatus.FOUND] and not os.getenv("CI"):
            pytest.skip("Could not contact RDF4J server")
    except Exception as exc:
        if os.getenv("CI"):
            pytest.fail(f"{exc}")
        else:
            pytest.skip(f"{exc}")
    repo = "test_repo"
    template = confpath() / "native_store_config_template.ttl"
    conf_bytes = config_bytes_from_template(template, {"repo": repo})
    service_cfg = ServiceConfig(repo, "http", url)
    current_repos = repos(service_cfg)

    assert "test_repo" not in [i.repo_id for i in current_repos]

    # protocol is added internally. Thus, skip from t_common.rdf4j_url
    client = new_repo(service_cfg, conf_bytes)
    current_repos = repos(service_cfg)
    assert "test_repo" in [i.repo_id for i in current_repos]

    client.delete_repo()
    current_repos = repos(service_cfg)
    assert "test_repo" not in [i.repo_id for i in current_repos]


def test_repos_with_auth(httpserver: HTTPServer):
    response_json = {
        "results": {
            "bindings": [
                {
                    "uri": {"value": "uri"},
                    "id": {"value": "id"},
                    "title": {"value": "title"},
                    "readable": {"value": "true"},
                    "writable": {"value": "false"},
                }
            ]
        }
    }

    matcher = HeaderValueMatcher({"authorization": lambda value, expect: expect == value})
    user, password = "user", "password"
    encoded_user_passwd = b64encode(bytes(f"{user}:{password}", "utf8")).decode("utf8")
    httpserver.expect_request(
        "/repositories",
        headers={"authorization": f"Basic {encoded_user_passwd}"},
        header_value_matcher=matcher,
    ).respond_with_json(response_json)
    url = httpserver.url_for("/repositories")

    matches = re.match(r"^([a-z]+):\/\/([a-z:0-9]+)", url)
    assert matches
    protocol, server = matches.groups()

    cfg = ServiceConfig("repo", server=server, protocol=protocol, user=user, passwd=password)
    repo_info = repos(cfg)

    expect = RepoInfo("uri", "id", "title", readable=True, writable=False)
    assert repo_info == [expect]


def test_update_prefixes():
    client = GraphDBClient(ServiceConfig(server="some-server", rest_api=RestApi.DIRECT_SPARQL_ENDPOINT))

    new_pref = {"eq": "http://eq"}
    client.update_prefixes(new_pref)
    assert client.prefixes["eq"] == new_pref["eq"]


def test_custom_headers():
    custom_headers = {"my_header": "my_header_value"}
    client = GraphDBClient(ServiceConfig(server="some-server", token=None), custom_headers)
    assert client.sparql.customHttpHeaders == custom_headers


class FixedResultSparqlWrapper(SPARQLWrapper):
    def __init__(self) -> None:
        super().__init__("http://fixed-result-endpoint")
        self.result = SparqlResultJsonFactory.build()

    def queryAndConvert(self) -> dict[str, Any]:  # noqa: N802
        return self.result.model_dump(mode="json")


def test_inject_subclassed_sparql_wrapper():
    wrapper = FixedResultSparqlWrapper()
    client = GraphDBClient(sparql_wrapper=wrapper)

    # Confirm that we can sucessfully run a query
    data = client.get_table("select * where {?s ?p ?o}")[0]
    assert set(data.columns) == set(wrapper.result.head.variables)


def test_xnodes(model: Model):
    dfs = all_data(model)
    con_nodes = dfs.connectivity_nodes
    ac_lines = dfs.ac_lines.assign(
        bidzone_1=lambda df: df["connectivity_node_1"].map(con_nodes["bidzone"]),
        bidzone_2=lambda df: df["connectivity_node_2"].map(con_nodes["bidzone"]),
    )
    xnode_branches = ac_lines[ac_lines["name"].str.contains("Xnode")]
    non_xnode_branches = ac_lines[~ac_lines["name"].str.contains("Xnode")]
    assert not xnode_branches.empty
    assert not xnode_branches["bidzone_1"].isna().any()
    assert not xnode_branches["bidzone_2"].isna().any()

    non_xnode_bidzones = set(non_xnode_branches["bidzone_1"]) | set(non_xnode_branches["bidzone_2"])
    xnode_bidzones = set(xnode_branches["bidzone_1"]) | set(xnode_branches["bidzone_2"])

    # Verify that there are no distinct bidzones among the xnodes such as EU, EU-ELSP-1 etc.
    assert xnode_bidzones.issubset(non_xnode_bidzones)



import os
from string import Template

import pandas as pd
import pytest

import tests.t_utils.common as t_common
import tests.t_utils.entsoe_models as t_entsoe


def check_service_available(test_model: t_common.ModelTest):
    if not test_model.model:
        if os.getenv("CI") and test_model.must_run_in_ci:
            pytest.fail("Service should always be available in CI")
        else:
            pytest.skip("Require access to external Graph service")


@pytest.mark.slow
@pytest.mark.parametrize("test_model", t_entsoe.smallgrid_models())
def test_eq_query(test_model: t_common.ModelTest):
    check_service_available(test_model)
    query_template = Template(
        "PREFIX rdf:<${rdf}>\nPREFIX cim:<${cim}>\n"
        "SELECT ?voltage {?s rdf:type cim:BaseVoltage; cim:BaseVoltage.nominalVoltage ?voltage}"
    )

    model = test_model.model
    assert model
    query = model.template_to_query(query_template)
    voltage = model.get_table_and_convert(query)
    voltage = voltage.sort_values("voltage").reset_index(drop=True)
    expect = pd.DataFrame({"voltage": [33.0, 132.0, 220.0]})
    pd.testing.assert_frame_equal(voltage, expect)


@pytest.mark.slow
@pytest.mark.parametrize("test_model", t_entsoe.smallgrid_models())
def test_tpsv_query(test_model: t_common.ModelTest):
    check_service_available(test_model)
    model = test_model.model
    assert model
    repo = model.config.system_state_repo
    query_template = Template(
        "PREFIX rdf:<${rdf}>\nPREFIX cim:<${cim}>\n"
        f"select * where {{service <{repo}> {{?s rdf:type cim:TopologicalNode}}}}"
    )
    query = model.template_to_query(query_template)
    topological_nodes = model.get_table_and_convert(query)
    assert topological_nodes.shape == (115, 1)



"""Provide retry call back."""

import logging
from collections.abc import Callable

from tenacity import RetryCallState

from cimsparql.utils import query_name

logger = logging.getLogger()


class RetryCallback:
    def __init__(self) -> None:
        self.query_name = ""

    def pre_call(self, query: str) -> None:
        try:
            self.query_name = query_name(query)
        except Exception as exc:
            logger.warning("pre_call: %s", exc)

    def before(self, _: RetryCallState) -> None: ...

    def after(self, retry_state: RetryCallState) -> None:
        if outcome := retry_state.outcome:
            logger.info(
                "Query: %s: Attempt %s ended with: %s",
                self.query_name,
                retry_state.attempt_number,
                outcome.exception(),
            )


RetryCallbackFactory = Callable[[], RetryCallback]



"""Adaption functionality used to modify test cases."""

from __future__ import annotations

import hashlib
import re
import uuid
from contextlib import suppress
from typing import TYPE_CHECKING

from rdflib import Graph
from rdflib.graph import Dataset
from rdflib.namespace import RDF, XSD
from rdflib.term import BNode, Literal, Node, URIRef

from cimsparql.graphdb import default_namespaces

if TYPE_CHECKING:
    from collections.abc import Iterable
    from pathlib import Path


class XmlModelAdaptor:
    eq_predicate = "http://entsoe.eu/CIM/EquipmentCore/3/1"

    def __init__(self, filenames: Iterable[Path]) -> None:
        self.graph = Dataset()
        self.ns = default_namespaces()
        for prefix, value in self.ns.items():
            self.graph.bind(prefix, value)

        for filename in filenames:
            profile = filename.stem.rpartition("/")[-1]
            uri = URIRef(f"http://cimsparql/xml-adpator/{profile}")
            graph = Graph(identifier=uri).parse(filename, publicID="http://cim")
            self.graph.addN([(s, p, o, graph) for s, p, o in graph])

    def namespaces(self) -> dict[str, str]:
        return self.ns | {str(prefix): str(name) for prefix, name in self.graph.namespaces()}

    def graphs(self) -> set[str]:
        return {graph for _, _, _, graph in self.graph.quads() if graph}

    @classmethod
    def from_folder(cls, folder: Path) -> XmlModelAdaptor:
        return XmlModelAdaptor(list(folder.glob("*.xml")))

    def add_mrid(self) -> None:
        """Add cim:IdentifiedObject.mRID if not present."""
        ns = self.namespaces()
        identified_obj_mrid = URIRef(f"{ns['cim']}IdentifiedObject.mRID")
        for ctx in self.graph.contexts():
            subjects = {s for s, _, _ in ctx}
            for s in subjects:
                mrid_str = str(s).rpartition("#_")[-1]
                mrid = mrid_str if is_uuid(mrid_str) else generate_uuid(mrid_str)
                ctx.add((s, identified_obj_mrid, Literal(mrid)))

    def add_market_code_to_non_conform_load(self) -> None:
        updated = set[Node]()
        for load, _, _, ctx in self.graph.quads((None, RDF.type, URIRef(self.ns["cim"] + "EnergyConsumer"), None)):
            if load in updated:
                continue

            if "EQ" not in str(ctx):
                continue
            updated.add(load)
            assert ctx

            current_graph = self.graph.get_graph(ctx)
            assert current_graph

            load_group = BNode()
            schedule_resource = BNode()
            current_graph.addN(
                [
                    (load, URIRef(self.ns["cim"] + "NonConformLoad.LoadGroup"), load_group, current_graph),
                    (
                        load_group,
                        URIRef(self.ns["cim"] + "IdentifiedObject.name"),
                        Literal("created-group", datatype=XSD.string),
                        current_graph,
                    ),
                    (
                        load_group,
                        URIRef(self.ns["SN"] + "NonConformLoadGroup.ScheduleResource"),
                        schedule_resource,
                        current_graph,
                    ),
                    (
                        schedule_resource,
                        URIRef(self.ns["SN"] + "ScheduleResource.marketCode"),
                        Literal("market001", datatype=XSD.string),
                        current_graph,
                    ),
                ]
            )

    def adapt(self, eq_uri: str) -> None:
        self.add_zero_sv_power_flow()
        self.add_zero_sv_injection()
        self.add_generating_unit()
        self.add_market_code_to_non_conform_load()
        self.add_mrid()
        self.add_dtypes()
        self.add_internal_eq_link(eq_uri)
        self.add_eic_code()
        self.add_network_analysis_enable()

    def add_zero_sv_power_flow(self) -> None:
        sv_graph = Graph(identifier=URIRef("http://cimsparql/xml-adpator/SV"))
        sv_power_flow_terminal = URIRef(self.ns["cim"] + "SvPowerFlow.Terminal")

        zero_sv_flow_added = set()

        for terminal, _, _, _ in self.graph.quads((None, RDF.type, URIRef(self.ns["cim"] + "Terminal"), None)):
            # Do not add SvPowerFlow if it already exists
            if any(self.graph.quads((None, sv_power_flow_terminal, terminal, None))) or terminal in zero_sv_flow_added:
                continue
            zero_sv_flow_added.add(terminal)
            power_flow = BNode()
            sv_graph.addN(
                (
                    (power_flow, sv_power_flow_terminal, terminal, sv_graph),
                    (power_flow, URIRef(self.ns["cim"] + "SvPowerFlow.p"), Literal(0.0), sv_graph),
                    (power_flow, URIRef(self.ns["cim"] + "SvPowerFlow.q"), Literal(0.0), sv_graph),
                )
            )
        self.graph.add_graph(sv_graph)

    def add_dtypes(self) -> None:
        fields = {
            "endNumber": XSD.integer,
            "sequenceNumber": XSD.integer,
            "phaseAngleClock": XSD.integer,
            "SvPowerFlow.p": XSD.float,
            ".open": XSD.boolean,
            ".connected": XSD.boolean,
            ".nominalVoltage": XSD.float,
        }
        for s, predicate, o, g in self.graph.quads():
            assert g
            current_graph = self.graph.get_graph(g)
            assert current_graph
            with suppress(StopIteration):
                f = next(f for f in fields if f in str(predicate))
                current_graph.remove((s, predicate, o))

                literal = Literal(str(o), datatype=fields[f])
                current_graph.add((s, predicate, literal))

    def tpsvssh_contexts(self) -> list[Graph]:
        return [ctx for ctx in self.graph.contexts() if any(token in str(ctx) for token in ("SSH", "TP", "SV"))]

    def eq_contexts(self) -> list[Graph]:
        return [ctx for ctx in self.graph.contexts() if any(token in str(ctx) for token in ("EQ", "GL"))]

    def nq_bytes(self, contexts: Iterable[Graph] | None = None) -> bytes:
        """Return the contexts as bytes. If contexts is None, the entire graph is exported."""
        if contexts is None:
            return self.graph.serialize(format="nquads", encoding="utf8")

        graph = Dataset()
        for ctx in contexts:
            graph.addN((s, p, o, ctx) for s, p, o in ctx)
        return graph.serialize(format="nquads", encoding="utf8")

    def add_internal_eq_link(self, eq_uri: str) -> None:
        # Insert in one SV graph
        ctx = next(c for c in self.tpsvssh_contexts() if "SV" in str(c))
        graph = self.graph.get_graph(ctx.identifier)
        assert graph
        graph.add((BNode(), URIRef(self.eq_predicate), URIRef(eq_uri)))

    def add_zero_sv_injection(self) -> None:
        tp_node_type = URIRef(self.ns["cim"] + "TopologicalNode")
        sv_inj_type = URIRef(self.ns["cim"] + "SvInjection")

        graph = Graph(identifier=URIRef("http://cimsparql/xml-adpator/SV-injection"))
        used_nodes = set()
        for s, _, _, _ in self.graph.quads((None, RDF.type, tp_node_type, None)):
            if s in used_nodes:
                continue
            used_nodes.add(s)
            sv_injection = BNode()
            graph.addN(
                (
                    (sv_injection, RDF.type, sv_inj_type, graph),
                    (sv_injection, URIRef(self.ns["cim"] + "SvInjection.TopologicalNode"), s, graph),
                    (
                        sv_injection,
                        URIRef(self.ns["cim"] + "SvInjection.pInjection"),
                        Literal(0.0, datatype=XSD.float),
                        graph,
                    ),
                )
            )
        self.graph.add_graph(graph)

    def add_eic_code(self) -> None:
        for substation, _, _, ctx in self.graph.quads((None, RDF.type, URIRef(self.ns["cim"] + "Substation"), None)):
            assert ctx
            current_graph = self.graph.get_graph(ctx)
            assert current_graph
            market_delivery_point = BNode()
            bidding_area = BNode()
            eic_code = Literal("10Y1001A1001A48H", datatype=XSD.string)
            current_graph.addN(
                (
                    (
                        substation,
                        URIRef(f"{self.ns['SN']}Substation.MarketDeliveryPoint"),
                        market_delivery_point,
                        current_graph,
                    ),
                    (
                        market_delivery_point,
                        URIRef(f"{self.ns['SN']}MarketDeliveryPoint.BiddingArea"),
                        bidding_area,
                        current_graph,
                    ),
                    (
                        bidding_area,
                        URIRef(f"{self.ns['entsoeSecretariat']}IdentifiedObject.energyIdentCodeEIC"),
                        eic_code,
                        current_graph,
                    ),
                )
            )

    def add_network_analysis_enable(self) -> None:
        used_equipment = set()
        for _, _, equipment, ctx in self.graph.quads(
            (None, URIRef(f"{self.ns['cim']}Terminal.ConductingEquipment"), None, None)
        ):
            if equipment in used_equipment:
                continue
            used_equipment.add(equipment)
            assert ctx
            graph = self.graph.get_graph(ctx)
            assert graph
            graph.add(
                (
                    equipment,
                    URIRef(f"{self.ns['SN']}Equipment.networkAnalysisEnable"),
                    Literal("true", datatype=XSD.boolean),
                ),
            )

    def add_generating_unit(self) -> None:
        updated = set()
        for sync_machine, _, _, ctx in self.graph.quads(
            (None, RDF.type, URIRef(self.ns["cim"] + "SynchronousMachine"), None)
        ):
            assert ctx
            if "EQ" not in str(ctx):
                continue
            if sync_machine in updated:
                continue

            generating_unit = BNode()
            schedule_resource = BNode()
            current_graph = self.graph.get_graph(ctx)
            assert current_graph
            updated.add(sync_machine)
            current_graph.addN(
                (
                    (
                        sync_machine,
                        URIRef(self.ns["cim"] + "SynchronousMachine.GeneratingUnit"),
                        generating_unit,
                        current_graph,
                    ),
                    (generating_unit, RDF.type, URIRef(self.ns["cim"] + "ThermalGeneratingUnit"), current_graph),
                    (
                        generating_unit,
                        URIRef(self.ns["cim"] + "IdentifiedObject.name"),
                        Literal("GeneratingUnit", datatype=XSD.string),
                        current_graph,
                    ),
                    (
                        generating_unit,
                        URIRef(self.ns["SN"] + "GeneratingUnit.ScheduleResource"),
                        schedule_resource,
                        current_graph,
                    ),
                    (schedule_resource, RDF.type, URIRef(self.ns["SN"] + "ScheduleResource"), current_graph),
                    (
                        schedule_resource,
                        URIRef(self.ns["SN"] + "ScheduleResource.marketCode"),
                        Literal("market001", datatype=XSD.string),
                        current_graph,
                    ),
                )
            )


def is_uuid(x: str) -> bool:
    return re.match("^[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}$", x) is not None


def generate_uuid(x: str) -> str:
    h = hashlib.md5(x.encode(), usedforsecurity=False)
    return str(uuid.UUID(hex=h.hexdigest()))
